import hashlib
import json
import logging
import os

from aiotstudio.datasource import search_df
from fastparquet import write, ParquetFile


class q_cacher:

    def __init__(self, cache_directory='parquet'):
        self.parquet_dir = cache_directory + '/'
        if not os.path.exists(self.parquet_dir):
            os.makedirs(self.parquet_dir)

    def _get_hash(self, query):
        hash_object = hashlib.sha1(json.dumps(query).encode('utf-8'))
        hex_dig = hash_object.hexdigest()
        return hex_dig

    def remove_cache_for(self, q, suffix=''):
        filename = self._create_fileName(q, suffix)
        os.remove(filename)

    def _is_query_cacheable(self, q):
        list_forbidden_keyword = ['thisMonth', 'now', 'sevenDaysAgo', 'today', 'yesterday', 'thirtyDaysAgo',
                                  'thisMonth',
                                  'lastMonth', 'thisQuarter', 'lastQuarter', 'thisYear', 'lastYear']
        for k in list_forbidden_keyword:
            if k.lower() in str(q).lower():
                return False
        return True

    def get_data_form_query(self, q, force=False):
        log = logging.getLogger()
        if self._is_data_cached(q) and not force:
            log.info('reading from cache')
            return self._get_data_from_cache(q)
        else:
            log.info('loading from smartobject')
            res = search_df(q)
        assert len(res) < 100000, str(q)
        if self._is_query_cacheable(q):
            self._save_df(res, q)
        return res

    def _create_fileName(self, q, suffix=''):
        if suffix != '':
            return self.parquet_dir + self._get_hash(q) + '_' + suffix + '.parq'
        return self.parquet_dir + self._get_hash(q) + '.parq'

    def _save_df(self, df, q, suffix=''):
        filename = self._create_fileName(q, suffix)
        write(filename, df, compression='SNAPPY')

    def _is_data_cached(self, q, suffix=''):
        filename = self._create_fileName(q, suffix)
        return os.path.isfile(filename)

    def _get_data_from_cache(self, q, suffix=''):
        filename = self._create_fileName(q, suffix)
        return ParquetFile(filename).to_pandas()



import sys
from os import environ

import pandas as pd
from pyspark import SparkContext, SparkConf


def init_spark(nb_cores='*'):
    python_exec = sys.executable

    environ['PYSPARK_PYTHON'] = python_exec
    environ['PYSPARK_DRIVER_PYTHON'] = python_exec
    environ['SPARK_YARN_USER_ENV'] = python_exec
    conf = SparkConf().setAppName("App")

    conf = (conf.setMaster('local[' + str(nb_cores) + ']')
            .set('spark.executor.memory', '4G')
            .set('spark.driver.memory', '45G')
            .set('spark.driver.maxResultSize', '10G'))
    sc = SparkContext(conf=conf)
    j_log = sc._jvm.org.apache.log4j
    j_log.LogManager.getRootLogger().setLevel(j_log.Level.OFF)
    return sc


def run_spark_job_concat_dfs(parameters_list, function_to_parallelize, nb_cores='*', sc=None):
    """
    This function should be used when the goal is to parallelilze a job across many parameters.
    The results are then concatenated into an dataframe
    Parameters to define:
    1. parameter_list: a list of the sets of arguments necessary to run the function to parallelize
    2. function to paralelize: function that must return an object to concatenate (most likely a dataframe). Its required arguments should be included in the parameter_list
    Default parameter:
    3. nb_cores: '*' is the default value meaning if not specified it will use all cores available.
    This function returns a concatenate dataframe
    """
    if sc is None:
        sc = init_spark(nb_cores)
    rdd = sc.parallelize(parameters_list)
    results = rdd.map(function_to_parallelize).reduce(lambda a, b: pd.concat([a, b], axis=0))
    return results


def run_spark_job_no_reduce(parameters_list, function_to_parallelize, nb_cores='*', sc=None):
    """
    This function should be used when the goal is to parallelilze a job across many parameters.
    The results are not reduced
    Parameters to define:
    1. parameter_list: a list of the sets of arguments necessary to run the function to parallelize
    2. function to paralelize: function that must return an object to concatenate (most likely a dataframe). Its required arguments should be included in the parameter_list
    Default parameter:
    3. nb_cores: '*' is the default value meaning if not specified it will use all cores available.
    This function returns a concatenate dataframe
    """
    if sc is None:
        sc = init_spark(nb_cores)
    rdd = sc.parallelize(parameters_list)
    rdd.map(function_to_parallelize).collect()
